# lib/pokan/peer.rb
module Pokan

  ##
  # Peer is a Entity that takes care of the peer semantics.
  # It controls the status, id and role attributes.
  class Peer < Entity

    def initialize
      super
      store(:role, 'peer', 0)
      store(:status, 'alive', 0)
    end

    def address=(address)
      @address = address
      self.id = "#{@address}:#{@udp_port}"  if defined?(@address) && defined?(@udp_port)
    end

    def address
      @address = id.split(':').first
    end

    def udp_port=(port)
      @udp_port = port
      self.id = "#{@address}:#{@udp_port}"  if defined?(@address) && defined?(@udp_port)
    end

    def udp_port
      @udp_port = id.split(':')[1]
    end

    def tcp_port=(port)
      store(:tcp_port, port)
    end

    def tcp_port
      value(:tcp_port)
    end

    ##
    # Stores the value and the timestamp for a given key.
    # If the key, which must be a symbol, already exists, it will be updated if
    # the timestamp is greater.
    # The timestamp defaults to Time.now.
    #
    # If the key is ':role', the values have to be 'peer' or 'seed'. In the same
    # way, if the key is ':status', the value have to be 'alive' or 'dead'.
    # Otherwise, the value will not be set.
    def store(key, value, timestamp = Time.now)
      case key
      when :role
        super(:role, value, timestamp)  if ['peer', 'seed'].include?(value)
      when :status
        super(:status, value, timestamp)  if ['alive', 'dead'].include?(value)
      else
        super(key, value, timestamp)
      end
    end

    ##
    # Changes the role of the peer. The role must be peer or seed. Otherwise, the role will stay as it was and no exception will be thrown.
    def role=(role)
      store(:role, role)
    end

    def role
      value(:role)
    end

    ##
    # Turns the peer into a seed.
    def act_as_seed
      store(:role, 'seed')
    end

    def seed?
      value(:role) == 'seed'
    end

    def peer?
      value(:role) == 'peer'
    end

    ##
    # Changes the status of the peer.
    # The status must be alive or dead. Otherwise, the role will stay as it was and no exception will be thrown.
    def status=(status)
      store(:status, status)
    end

    def status
      value(:status)
    end

    ##
    # Revives the peer
    def revive
      store(:status, 'alive')
    end

    ##
    # Kills the peer
    def kill
      store(:status, 'dead')
      self
    end

    def alive?
      status == 'alive'
    end

    def dead?
      status == 'dead'
    end

    ##
    # Gossips with the given seed and replicates its local storage.
    #
    # If the host or port is invalid, the local storage will continue empty.
    # === Usage
    #   peer = Pokan::Peer.new
    #   peer.address = '127.0.0.2'
    #   peer.port = '1234'
    #   peer.value(:key_not_present) # => nil
    #   peer.gossip_with('127.0.0.2', '8888')
    #   peer.value(:key_not_present) # => 'the value of the key'
    def sync_with(host, port)
      db = Connection.redis
      db.slaveof(host, port)
      sleep(3) #while db.info['master_sync_in_progress'].to_i == 0
      @seed = host
      db.slaveof('no', 'one')
    end
  end
end
